import re
import os
import sys
import mergedeep

from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory

from datamodel_code_generator import InputFileType, generate
from datamodel_code_generator import DataModelType

from runch._reader import str_to_dict, _SupportedFileType

from typing import Any, Literal, cast

__doc__ = """Usage: python -m runch <config_path> [config_name [config_ext]]
    Generate a model definition from a config file.

    config_path: path to your config file.
    config_name: controls generated variable name and class name.
    config_type: content type of your config file. Default is `yaml`.

    Example:
        python -m runch path/to/my_config.foo
        python -m runch path/to/my_config.foo chat_config
        python -m runch path/to/my_config.foo chat_config yaml"""


@dataclass
class FileNameInfo:
    name: str
    ext: str

    def __str__(self) -> str:
        if self.ext:
            return f"{self.name}.{self.ext}"
        else:
            return self.name


def parse_file_name(file_name: str) -> FileNameInfo:
    # is a path?
    if os.path.sep in file_name:
        raise ValueError(f"Invalid file name: {file_name}")

    name, ext = os.path.splitext(file_name)
    ext = ext[1:]

    return FileNameInfo(name=name, ext=ext)


def generate_model(
    config_path: str, file_type: _SupportedFileType, config_name: str | None = None
):
    if file_type not in ["yaml", "json", "toml"]:
        raise ValueError(f"Unsupported content type: {file_type}")

    config_file_name = os.path.basename(config_path)
    config_file_name_info = parse_file_name(config_file_name)

    if config_file_name_info.name.endswith(".example"):
        config_file_name_info.name = config_file_name_info.name[:-8]
        config_path = os.path.join(
            os.path.dirname(config_path), str(config_file_name_info)
        )

    example_config_name = ".".join(
        [config_file_name_info.name, "example", config_file_name_info.ext]
    )
    example_config_path = os.path.join(
        os.path.dirname(config_path), example_config_name
    )

    config: dict[Any, Any] = {}
    example_config: dict[Any, Any] = {}

    config_exists = False
    example_config_exists = False

    try:
        with open(config_path, "r") as f:
            config = str_to_dict(
                f.read(),
                file_type,
                filename=f.name,
            )
            config_exists = True
    except FileNotFoundError:
        pass

    try:
        with open(example_config_path, "r") as f:
            example_config = str_to_dict(
                f.read(),
                file_type,
                filename=f.name,
            )
            example_config_exists = True
    except FileNotFoundError:
        pass

    if not config_exists and not example_config_exists:
        raise FileNotFoundError(
            f"Neither {config_path} nor {example_config_path} exists"
        )

    merged_config = mergedeep.merge(
        example_config, config, strategy=mergedeep.Strategy.TYPESAFE_REPLACE
    )

    if config_file_name_info.ext != "":
        display_ext = "." + config_file_name_info.ext
    else:
        display_ext = ""

    if config_name is None:
        if len(config_file_name_info.name) == 0:
            raise ValueError(
                "config_name is required as we can't infer it from the provided file name"
            )

        initial_inferred_config_name = config_file_name_info.name

        allowed_leading_chars = re.compile(r"[a-zA-Z_]")
        allowed_chars = re.compile(r"[a-zA-Z0-9_]")

        inferred_config_name = initial_inferred_config_name

        if len(inferred_config_name) == 1 and not allowed_leading_chars.match(
            inferred_config_name
        ):
            raise ValueError(
                f"Invalid inferred config_name: {initial_inferred_config_name}"
            )

        if len(inferred_config_name) > 1:
            # first, select the first valid leading char
            while len(inferred_config_name) > 0 and not allowed_leading_chars.match(
                inferred_config_name[0]
            ):
                inferred_config_name = inferred_config_name[1:]

            if len(inferred_config_name) == 0:
                raise ValueError(
                    f"Invalid inferred config_name: {initial_inferred_config_name}"
                )

            # then, remove all invalid following chars
            curr_checking_index = 1

            while curr_checking_index < len(inferred_config_name):
                has_stripped_chars = False
                while curr_checking_index < len(
                    inferred_config_name
                ) and not allowed_chars.match(
                    inferred_config_name[curr_checking_index]
                ):
                    inferred_config_name = (
                        inferred_config_name[:curr_checking_index]
                        + inferred_config_name[curr_checking_index + 1 :]
                    )
                    has_stripped_chars = True

                if (
                    has_stripped_chars
                    and not inferred_config_name[curr_checking_index - 1] == "_"
                ):
                    # add underscore. data-model-generator will convert snake_case to CamelCase
                    # when generating class names, so we don't need to worry about it
                    inferred_config_name = (
                        inferred_config_name[:curr_checking_index]
                        + "_"
                        + inferred_config_name[curr_checking_index:]
                    )

                if curr_checking_index < len(inferred_config_name):
                    curr_checking_index += 1

        config_name = inferred_config_name.rstrip("_")

    config_display_name = config_file_name_info.name + "{.example,}" + display_ext

    header = f"# Generated from {config_display_name} by runch"
    header += "\n# Please be aware that `float` fields might be annotated as `int` due to the lack of type info in the config."

    with TemporaryDirectory() as temporary_directory_name:
        temporary_directory = Path(temporary_directory_name)
        rand = os.urandom(1).hex()
        output = Path(temporary_directory / f"model_{rand}.py")

        generate(
            merged_config,
            input_file_type=InputFileType.Dict,
            input_filename="placeholder",
            output=output,
            output_model_type=DataModelType.PydanticV2BaseModel,
            custom_file_header=header,
            custom_formatters=["runch.script.custom_formatter"],
            custom_formatters_kwargs={
                "config_file_ext": config_file_name_info.ext,
                "config_name": config_name,
                "config_path": config_path,
                "config_type": file_type,
            },
            snake_case_field=True,
        )
        model: str = output.read_text()

    return model


if __name__ == "__main__":
    # TODO: move to argparse
    if len(sys.argv) < 2:
        print(
            __doc__,
            file=sys.stderr,
        )
        sys.exit(1)

    config_path = sys.argv[1]
    config_name = None
    config_type: _SupportedFileType = "yaml"

    if len(sys.argv) == 3:
        config_name = sys.argv[2]
    elif len(sys.argv) == 4:
        config_name = sys.argv[2]
        config_type = cast(_SupportedFileType, sys.argv[3])
    elif len(sys.argv) > 4:
        print(
            __doc__,
            file=sys.stderr,
        )
        sys.exit(1)

    model = generate_model(config_path, config_type, config_name=config_name)
    print(model)



from runch._reader import (
    FeatureConfig,
    RunchConfigReader,
    RunchAsyncCustomConfigReader,
    require_lazy_runch_configs,
)
from runch.runch import Runch, RunchModel

__all__ = [
    "Runch",
    "RunchModel",
    "RunchConfigReader",
    "RunchAsyncCustomConfigReader",
    "FeatureConfig",
    "require_lazy_runch_configs",
]



import sys
import typing

from typing import (
    Any,
    Generic,
    ParamSpec,
    Type,
    TypeVar,
    TypeVarTuple,
    get_args,
    get_origin,
)

_typing_3_7 = False
try:
    from typing import ForwardRef as _

    _typing_3_7 = True
except:
    pass

T = TypeVar("T")


def is_generic(tp: Type[Any]):
    """adapted from pytypes.get_orig_class: Check if a type is a generic type."""
    try:
        return isinstance(tp, typing.GenericMeta)  # type: ignore
    except AttributeError:
        try:
            return issubclass(tp, typing.Generic)
        except AttributeError:
            return False
        except TypeError:
            return isinstance(tp, typing._GenericAlias)  # type: ignore


def unwrap[T](x: T | None) -> T:
    if x is None:
        raise Exception("unwrap: arg should not be None")
    return x


def currentframe():
    """adapted from pytypes.get_orig_class: Return the frame of the caller."""
    if hasattr(sys, "_getframe"):
        return sys._getframe(1)  # type: ignore
    raise ValueError("No frame information")


def get_orig_class(obj: Any, default_to__class__: bool = False) -> Any:
    """adapted from pytypes.get_orig_class:

    Robust way to access `obj.__orig_class__`. Compared to a direct access this has the
    following advantages:
    1) It works around https://github.com/python/typing/issues/658.
    2) It prevents infinite recursion when wrapping a method (`obj` is `self` or `cls`) and either
       - the object's class defines `__getattribute__`
       or
       - the object has no `__orig_class__` attribute and the object's class defines `__getattr__`.
       See discussion at https://github.com/Stewori/pytypes/pull/53.
    If `default_to__class__` is `True` it returns `obj.__class__` as final fallback.
    Otherwise, `AttributeError` is raised in failure case (default behavior).
    """
    try:
        # See https://github.com/Stewori/pytypes/pull/53:
        # Returns  `obj.__orig_class__` protecting from infinite recursion in `__getattr[ibute]__`
        # wrapped in a `checker_tp`.
        # (See `checker_tp` in `typechecker._typeinspect_func for context)
        # Necessary if:
        # - we're wrapping a method (`obj` is `self`/`cls`) and either
        #     - the object's class defines __getattribute__
        # or
        #     - the object doesn't have an `__orig_class__` attribute
        #       and the object's class defines __getattr__.
        # In such a situation, `parent_class = obj.__orig_class__`
        # would call `__getattr[ibute]__`. But that method is wrapped in a `checker_tp` too,
        # so then we'd go into the wrapped `__getattr[ibute]__` and do
        # `parent_class = obj.__orig_class__`, which would call `__getattr[ibute]__`
        # again, and so on. So to bypass `__getattr[ibute]__` we do this:
        return object.__getattribute__(obj, "__orig_class__")
    except AttributeError:
        if sys.version_info.major >= 3:
            cls = object.__getattribute__(obj, "__class__")
        else:
            # Python 2 may return instance objects from object.__getattribute__.
            cls = obj.__class__
        if _typing_3_7 and is_generic(cls):
            # Workaround for https://github.com/python/typing/issues/658
            # Searching from index 2 is sufficient: At 0 is get_orig_class, at 1 is the caller.
            # We assume the caller is not typing._GenericAlias.__call__ which we are after.
            frame = unwrap(currentframe().f_back).f_back
            try:
                while frame:
                    try:
                        res = frame.f_locals["self"]
                        if res.__origin__ is cls:
                            return res
                    except (KeyError, AttributeError):
                        frame = frame.f_back
            finally:
                del frame

        if default_to__class__:
            return cls  # Fallback
        raise


def get_generic_arg_kv_map(type_: Type[T]) -> dict[TypeVar | TypeVarTuple, Type[Any]]:
    type_origin = get_origin(type_) or type_
    generic_arg_kv_map: dict[TypeVar | TypeVarTuple, Type[Any]] = {}

    # this check is equivalent to `_is_generic_type(type_)`
    if type_origin != type_:
        generic_arg_types = get_args(type_)

        # we can't just use issubclass(type_origin, Generic) here,
        # because both `class X[T]:` and `class X(Generic[T])` are subclass of Generic
        if issubclass(type_origin, Generic):
            if len(type_origin.__type_params__) == 0:
                # class X(Generic[T, U]):
                generic_arg_vars = get_args(type_origin.__orig_bases__[0])  # type: ignore
            else:
                # class X[T, U]:
                generic_arg_vars = type_origin.__type_params__

            if len(generic_arg_vars) != len(generic_arg_types):
                raise ValueError(
                    f"get_generic_arg_kv_map: {type_origin=} is malformed, typevar arg count {len(generic_arg_vars)=} != provided arg count {len(generic_arg_types)=}"
                )

            for i in range(len(generic_arg_vars)):
                var = generic_arg_vars[i]
                if isinstance(var, ParamSpec):
                    # we don't include ParamSpec in the returned map
                    continue
                if not isinstance(var, TypeVar) and not isinstance(var, TypeVarTuple):
                    raise ValueError(
                        f"get_generic_arg_kv_map: {type_origin=} has non-TypeVar generic arg {var}"
                    )
                generic_arg_kv_map[var] = generic_arg_types[i]

    return generic_arg_kv_map



from __future__ import annotations

import yaml
import pydantic

from pydantic_core import from_json

from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Type,
    TypeAlias,
    TypeVar,
    cast,
)
from typing_extensions import deprecated

__all__ = ["RunchModel", "Runch"]

IncEx: TypeAlias = set[int] | set[str] | dict[int, Any] | dict[str, Any] | None
T = TypeVar("T")


class RunchModel(pydantic.BaseModel):
    """`RunchModel` is a pydantic `BaseModel` with extra configurations. All config models should inherit from this class.

    Note: Please avoid using `enum.Enum` as a field type in your `RunchModel`. Pydantic has limited support for enums.
    """

    model_config = pydantic.ConfigDict(
        extra="ignore",
        strict=True,
        protected_namespaces=(),
        use_enum_values=True,
    )

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    def get(self, key: str, default: Any = None) -> Any:
        if key not in self.model_fields:
            return default

        try:
            return self.__getattribute__(key)
        except AttributeError:
            return default


class Runch[C: RunchModel](pydantic.RootModel[C]):
    """Runch Config Class

    Args:
        Generic TypeVar `T`: a `RunchModel` defining the schema of a config.

    Raises:
        `ValidationError`: when provided args or kwargs do not match your schema.

    Exmaple:
        ```python
        class Test(RunchModel):
            a: int
            b: str

        config = Runch[Test]({"a": 1, "b": "c"}).config
        print(config.a)
        ```
    """

    if TYPE_CHECKING:
        # User needs to use .config and should not access the .root attribute directly.
        root: None
    else:
        # we need to expose this attribute to pyright for run-time reflection
        root: C

    def __init__(self, *args: Any, **kwargs: Any):
        self.__init_args = args
        self.__init_kwargs = kwargs
        super().__init__(  # pyright: ignore[reportUnknownMemberType]
            *self.__init_args, **self.__init_kwargs
        )

    @property
    def config(self) -> C:
        return cast(C, self.root)

    @config.setter
    def config(self, value: C) -> None:
        self.root = value  # pyright: ignore

    @classmethod
    def signature(cls) -> Type[C]:
        # this is a workaround for dynamic generated generic class `get_args(Runch[type_])` is ().
        # We have to use dynamic generated generic class Runch[type_] because Runch[C] causes the
        # actual type of `C` to be lost.
        # Conclusion: ConfigReader[C] -> type_ = get_generic_arg_kv_map(cls)[C] -> Runch[type_] and
        # Runch[type_].signature() returns type_. This also works for static type Runch[X] where X
        # is a statically defined RunchModel.
        return cls.__signature__.parameters["root"].annotation

    def __getitem__(self, key: str) -> Any:
        return self.config.__getattribute__(key)

    def __setitem__(self, key: str, value: Any) -> None:
        raise NotImplementedError(
            "Runch config is immutable. Use `update` method to create a new config in-place."
        )

    def __iter__(self):
        return iter(self.config)

    def __len__(self):
        return len(self.config.model_fields)

    def __contains__(self, key: str) -> bool:
        return key in self.config.model_fields

    def __repr__(self):
        sig = self.signature()

        return f"<Runch of {sig.__qualname__}>"

    def __str__(self):
        return repr(self)

    @deprecated("use __getitem__ instead")
    def get(self, key: str, default: Any = None) -> Any:
        try:
            self.config.__getattribute__(key)
        except AttributeError:
            return default

    if TYPE_CHECKING:

        def toJSON(
            self,
            *,
            indent: int | None = None,
            include: IncEx = None,
            exclude: IncEx = None,
            by_alias: bool = False,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,
            round_trip: bool = False,
            warnings: bool = True,
            serialize_as_any: bool = False,
            **dumps_kwargs: Any,
        ) -> str: ...

        to_json = toJSON

    else:

        def toJSON(self, *args: Any, **kwargs: Any) -> str:
            return self.config.model_dump_json(
                *args,
                **kwargs,
            )

        to_json = toJSON

    @classmethod
    def fromJSON(cls, json_str: str):
        type_ = cls.signature()
        return Runch[type_](from_json(json_str))

    @classmethod
    def from_json(cls, json_str: str):
        return cls.fromJSON(json_str)

    if TYPE_CHECKING:

        def toYAML(
            self,
            *,
            allow_unicode: bool = True,
            default_style: str | None = None,
            default_flow_style: bool | None = False,
            canonical: bool | None = None,
            indent: int | None = None,
            width: int | float | None = None,
            line_break: str | None = None,
            explicit_start: bool | None = None,
            explicit_end: bool | None = None,
            version: tuple[int, int] | None = None,
            tags: Mapping[str, str] | None = None,
            sort_keys: bool = True,
            **dumps_kwargs: Any,
        ) -> str: ...

        to_yaml = toYAML

    else:

        def toYAML(
            self,
            *,
            allow_unicode: bool = True,
            **kwargs: Any,
        ) -> str:
            return yaml.safe_dump(
                data=self.toJSON(),
                stream=None,  # make sure we return a string instead of None
                encoding=None,  # make sure we return a string instead of bytes
                allow_unicode=allow_unicode,  # default is changed to True
                **kwargs,
            )

        to_yaml = toYAML

    @classmethod
    def fromYAML(cls, yaml_str: str):
        type_ = cls.signature()
        return Runch[type_](yaml.safe_load(yaml_str))

    @classmethod
    def from_yaml(cls, yaml_str: str):
        return cls.fromYAML(yaml_str)

    if TYPE_CHECKING:

        def toDict(
            self,
            *,
            mode: str | Literal["json", "python"] = "python",
            include: IncEx = None,
            exclude: IncEx = None,
            context: Any | None = None,
            by_alias: bool = False,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,
            round_trip: bool = False,
            warnings: bool | Literal["none", "warn", "error"] = True,
            serialize_as_any: bool = False,
            **dumps_kwargs: Any,
        ) -> dict[str, Any]: ...

        to_dict = toDict

    else:

        def toDict(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
            return self.config.model_dump(*args, **kwargs)

        to_dict = toDict

    @classmethod
    def fromDict(
        cls,
        dict_obj: dict[Any, Any],
    ):
        type_ = cls.signature()
        return Runch[type_](dict_obj)

    @classmethod
    def from_dict(cls, dict_obj: dict[Any, Any]):
        return cls.fromDict(dict_obj)

    def update(self, new_runch: Runch[C]):
        """Update (replace) the current config with a new config."""

        self.config = new_runch.config

    # @classmethod
    # def from_model(cls, model: RunchModel):
    #     """Create a new Runch instance from a pydantic model."""
    #     type_ = cls.signature()
    #     return Runch[type_](model.model_dump())



# pyright: basic
from __future__ import annotations

import asyncio
import json
import os
import threading
import tomllib
import weakref
import yaml
import mergedeep

from datetime import datetime
from hashlib import blake2b
from io import StringIO

from .runch import Runch, RunchModel
from ._type_utils import get_orig_class, get_generic_arg_kv_map

from collections.abc import Callable, Coroutine, Mapping, MutableMapping
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    NamedTuple,
    Self,
    Type,
    TypeVar,
    TypeAlias,
    TypeVarTuple,
    TypedDict,
    assert_type,
    cast,
    get_args,
)


M = TypeVar("M", bound=RunchModel)
FeatureKey: TypeAlias = Literal["watch_update", "merge_example"]
AsyncFeatureKey: TypeAlias = Literal["watch_update"]
SupportedFileType: TypeAlias = Literal["yaml", "json", "toml"]

_UserCustomFileType: TypeAlias = Literal["_user_custom"]
_USER_CUSTOM_FILE_TYPE: _UserCustomFileType = "_user_custom"

_RUNCH_DEFAULT_CONFIG_DIR = os.environ.get(
    "RUNCH_CONFIG_DIR", os.path.join(os.getcwd(), "etc")
)

_SupportedFileType: TypeAlias = SupportedFileType | _UserCustomFileType

_normalized_supported_file_types: set[str] = set(get_args(SupportedFileType))


class FeatureConfig(TypedDict):
    enabled: bool
    args: dict[str, Any]


def str_to_dict(
    s: str,
    filetype: _SupportedFileType,
    *,
    filename: str,
    custom_loader: Callable[[str], dict[Any, Any]] | None = None,
) -> dict[Any, Any]:
    if custom_loader is not None:
        # custom loader is provided, we don't care about the file extension. we return whatever the custom loader returns
        # without checking if it is a dict. it is the user's responsibility to ensure the returned data is what they want
        return custom_loader(s)

    if filetype == "yaml":
        f = StringIO(s)
        config_dict = yaml.safe_load(f)
        # yaml.safe_load may return None if the file is empty, we should make an empty config be a valid config
        if config_dict is None:
            config_dict = {}
        if not isinstance(config_dict, dict):
            raise TypeError(
                f"Invalid config format: {filename} type={type(config_dict)}, expecting a dict"
            )
        return cast(dict[Any, Any], config_dict)
    elif filetype == "json":
        config_dict = json.loads(s)
        # we may got a list or even a string / number from json.load, and runtime type checking for these is not supported
        if not isinstance(config_dict, dict):
            raise TypeError(
                f"Invalid config format: {filename} type={type(config_dict)}, expecting a dict"
            )
        return cast(dict[Any, Any], config_dict)
    elif filetype == "toml":
        # According to tomllib docs, a whole toml document is always parsed into a dict
        return tomllib.loads(s)
    elif filetype == _USER_CUSTOM_FILE_TYPE:
        if custom_loader is None:
            raise ValueError(
                "custom_loader must be provided when reading configs from user custom file type"
            )
        return custom_loader(s)
    else:
        # dead code just for completeness
        assert_type("dead code reached", None)
        raise ValueError(f"Unsupported file type: {filetype}")


class FileNameInfo(NamedTuple):
    name: str
    ext: str


def parse_file_name(file_name: str) -> FileNameInfo:
    # is a path?
    if os.path.sep in file_name:
        raise ValueError(f"Invalid file name: {file_name}")

    name, ext = os.path.splitext(file_name)
    ext = ext[1:]

    return FileNameInfo(name=name, ext=ext)


class VersionedConfig(NamedTuple):
    config: dict[Any, Any]
    version: bytes


def read_config(
    config_name: str,
    config_dir: str,
    config_type: SupportedFileType | str = "yaml",
    config_encoding: str = "utf-8",
    *,
    custom_loader: Callable[[str], dict[Any, Any]] | None = None,
    should_merge_example: bool = False,
) -> VersionedConfig:
    real_config_path = os.path.join(config_dir, config_name)

    config_file_name_info = parse_file_name(config_name)

    example_config_name = ".".join(
        [config_file_name_info.name, "example", config_file_name_info.ext]
    )
    example_config_path = os.path.join(config_dir, example_config_name)

    if not should_merge_example:
        with open(real_config_path, "rb") as f:
            content_bytes = f.read()
            content = content_bytes.decode(encoding=config_encoding)
            version = blake2b(content_bytes).digest()
            if config_type in _normalized_supported_file_types:
                return VersionedConfig(
                    config=str_to_dict(
                        content,
                        cast(_SupportedFileType, config_type),
                        filename=f.name,
                        custom_loader=custom_loader,
                    ),
                    version=version,
                )
            else:
                return VersionedConfig(
                    config=str_to_dict(
                        content,
                        _USER_CUSTOM_FILE_TYPE,
                        filename=f.name,
                        custom_loader=custom_loader,
                    ),
                    version=version,
                )

    real_config: dict[Any, Any] = {}
    example_config: dict[Any, Any] = {}

    real_config_exists = False
    example_config_exists = False

    raw_configs_bytes = b""

    try:
        with open(real_config_path, "rb") as f:
            content_bytes = f.read()
            raw_configs_bytes += content_bytes
            real_config = str_to_dict(
                content_bytes.decode(encoding=config_encoding),
                cast(_SupportedFileType, config_type),
                filename=f.name,
                custom_loader=custom_loader,
            )
            real_config_exists = True
    except FileNotFoundError:
        pass

    try:
        with open(example_config_path, "rb") as f:
            content_bytes = f.read()
            raw_configs_bytes += content_bytes
            example_config = str_to_dict(
                content_bytes.decode(encoding=config_encoding),
                cast(_SupportedFileType, config_type),
                filename=f.name,
                custom_loader=custom_loader,
            )
            example_config_exists = True
    except FileNotFoundError:
        pass

    if not real_config_exists and not example_config_exists:
        raise FileNotFoundError(
            f"Neither {real_config_path} nor {example_config_path} exists"
        )

    merged_config = cast(
        dict[str, Any],
        mergedeep.merge(
            example_config, real_config, strategy=mergedeep.Strategy.TYPESAFE_REPLACE
        ),
    )
    return VersionedConfig(
        config=merged_config, version=blake2b(raw_configs_bytes).digest()
    )


_CONFIG_READER_DEFAULT_FEATURES: dict[FeatureKey, FeatureConfig] = {
    "watch_update": FeatureConfig(enabled=False, args={}),
    "merge_example": FeatureConfig(enabled=False, args={}),
}

_CUSTOM_A_CONFIG_READER_DEFAULT_FEATURES: dict[AsyncFeatureKey, FeatureConfig] = {
    "watch_update": FeatureConfig(enabled=False, args={}),
}

T = TypeVar("T")
U = TypeVarTuple("U")


def _run_sync_in_background(func: Callable[[*U], T], *args: *U) -> None:
    def thread_entry():
        func(*args)

    thread = threading.Thread(target=thread_entry, daemon=True)
    thread.start()


class RunchConfigReader[C: RunchModel]:
    _config_name: str
    _config_dir: str
    _config_type: SupportedFileType | str
    _config_encoding: str

    _config_schema: Type[C]
    _config: Runch[C] | None
    _config_version: bytes | None
    _config_updated_at: datetime | None

    _custom_config_parser: Callable[[str], dict[Any, Any]] | None

    _features: MutableMapping[FeatureKey, FeatureConfig]
    _auto_update_started: bool

    if TYPE_CHECKING:
        # this attribute is defined on the Generic class, simply adding it here to avoid pyright error
        __orig_class__: Type[RunchConfigReader[C]]

    def __init__(
        self,
        config_name: str,
        config_dir: str = _RUNCH_DEFAULT_CONFIG_DIR,
        config_type: SupportedFileType | str = "yaml",
        config_encoding: str = "utf-8",
        *,
        custom_config_parser: Callable[[str], dict[Any, Any]] | None = None,
        features: Mapping[FeatureKey, FeatureConfig] | None = None,
    ):
        self._config = None
        self._config_name = config_name
        self._config_dir = config_dir
        self._config_type = config_type
        self._config_encoding = config_encoding
        self._config_schema = get_generic_arg_kv_map(get_orig_class(self))[C]
        self._custom_config_parser = custom_config_parser
        self._features = {}
        self._config_updated_at = None
        self._config_version = None
        self._auto_update_started = False

        if features is not None:
            for feature, value in features.items():
                self.set_feature(feature, value)

        # make sure _features is always fully initialized
        for feature, value in _CONFIG_READER_DEFAULT_FEATURES.items():
            if feature not in self._features:
                self.set_feature(feature, value)

    def read(self) -> Runch[C]:
        if self._config is not None:
            return self._config

        type_ = self._config_schema
        versioned_config = read_config(
            self._config_name,
            self._config_dir,
            self._config_type,
            self._config_encoding,
            custom_loader=self._custom_config_parser,
            should_merge_example=self._features["merge_example"]["enabled"],
        )

        if (
            self._config_version == versioned_config.version
            and self._config is not None
        ):
            # no need to update the config
            return self._config

        self._config = Runch[type_].fromDict(versioned_config.config)
        self._config_version = versioned_config.version
        self._config_updated_at = datetime.now()

        return self._config

    def update(
        self,
        *,
        overwrite_uninitialized: bool = False,
        on_error: Literal["ignore", "raise"] = "ignore",
    ):
        """This function will try to update the config base on the latest config file.

        Args:
            overwrite_uninitialized (bool, optional): _If set to False, then update() will become a noop before read() is ever called. This can have side effects when used together with `read_lazy()`._
            Defaults to `False`.

            on_error ("raise" | "ignore" , optional): _Specifies the behavior when errors occurred during an attempt to update this config._
            Defaults to `"ignore"`, which means ignore the error and keep the old config.
        """

        if self._config is None:
            if self._config_updated_at is not None:
                raise RuntimeError(
                    "update: _config is None but _config_updated_at is not None. This is a bug."
                )
            if overwrite_uninitialized:
                # force update even if the config is not initialized due to lazy load
                try:
                    self.read()
                except Exception:
                    if on_error == "raise":
                        raise

            return

        type_ = self._config_schema
        versioned_config = read_config(
            self._config_name,
            self._config_dir,
            self._config_type,
            self._config_encoding,
            custom_loader=self._custom_config_parser,
            should_merge_example=self._features["merge_example"]["enabled"],
        )

        try:
            new_config = Runch[type_].fromDict(versioned_config.config)
        except Exception:
            if on_error == "raise":
                raise
            else:
                return

        if self._config_version != versioned_config.version:
            self._config.update(new_config)
            self._config_version = versioned_config.version
            self._config_updated_at = datetime.now()

    def read_lazy(self) -> Runch[C]:
        """
        Returns a lazy proxy object that will not read the config until any attribute is accessed

        Type checks are also postponed
        """
        type_ = self._config_schema
        that = self

        class Proxy:
            def __getattribute__(self, name: str) -> Any:
                if that._config is not None:
                    return that._config.__getattribute__(name)

                versioned_config = read_config(
                    that._config_name,
                    that._config_dir,
                    that._config_type,
                    that._config_encoding,
                    custom_loader=that._custom_config_parser,
                    should_merge_example=that._features["merge_example"]["enabled"],
                )

                # no need to check if config_version matches, because we are sure it is None
                that._config = Runch[type_].fromDict(versioned_config.config)
                that._config_version = versioned_config.version
                that._config_updated_at = datetime.now()

                return that._config.__getattribute__(name)

            def __repr__(self) -> str:
                if that._config is not None:
                    return f"<Proxy of {repr(that._config)}>"
                return f"<Proxy of {repr(Runch[type_])} (not evaluated yet)>"

        return Proxy()  # pyright: ignore[reportReturnType]

    def set_feature(self, feature: FeatureKey, feature_config: FeatureConfig) -> Self:
        """Set a feature's configuration. See **Features** section for available features.

        Args:
            feature (FeatureKey): The feature's key
            value (FeatureConfig): The feature's configuration.
              - enabled: Whether the feature is enabled
              - args: The arguments for the feature

        ## Features
        - feature="watch_update": Automatically update the config every `n` seconds.
            - args: `{"update_interval": n, "on_update_error": "ignore" | "raise"}`, where `update_interval` is always required.
        - feature="merge_example": Merge the example config with the actual config. This is useful for development.
            - args: `{}`.
        """
        self._features[feature] = feature_config

        if feature == "watch_update" and feature_config["enabled"]:
            self._start_auto_update()

        return self

    def enable_feature(self, feature: FeatureKey, args: dict[str, Any]) -> Self:
        return self.set_feature(feature, FeatureConfig(enabled=True, args=args))

    def disable_feature(self, feature: FeatureKey) -> Self:
        return self.set_feature(feature, FeatureConfig(enabled=False, args={}))

    def _start_auto_update(self):
        if not self._features["watch_update"]["enabled"]:
            return

        if "update_interval" not in self._features["watch_update"]["args"]:
            raise ValueError("watch_update feature requires `update_interval` in args")

        if "on_update_error" in self._features["watch_update"]["args"]:
            on_update_error = self._features["watch_update"]["args"]["on_update_error"]
            if on_update_error not in ["ignore", "raise"]:
                raise ValueError('on_update_error must be either "ignore" or "raise"')
        else:
            # will fallback to default later
            on_update_error = None

        if self._auto_update_started:
            return

        self._auto_update_started = True

        # use weakref to avoid circular reference to `self`
        self_ref = weakref.ref(self)

        # watchdog will stop automatically if the feature is disabled / the reader is deleted
        async def a_watchdog():
            # NOTE: under very rare circumstances, the auto update watchdog thread may not be able to be recreated.
            # That is, if the user disables this feature, watchdog will quit in the next iteration. However, if the
            # user enables the feature again between the `break` of `while True` and `self._auto_updating = False`,
            # the new watchdog thread will not be created due to the `if self._auto_updating` check.
            # This is a very very rare edge case and is not quite fixable without adding a lock and some complexity.
            while True:
                self_ = self_ref()
                if self_ is None:
                    # reader is deleted
                    break

                await asyncio.sleep(
                    self_._features["watch_update"]["args"]["update_interval"]
                )

                if not self_._features["watch_update"]["enabled"]:
                    # auto update is turned off
                    break

                if on_update_error is not None:
                    self_.update(
                        overwrite_uninitialized=False, on_error=on_update_error
                    )
                else:
                    self_.update(overwrite_uninitialized=False)

                del self_

            self_ = self_ref()
            if self_ is not None:
                self_._auto_update_started = False

        def watchdog():
            asyncio.run(a_watchdog())

        _run_sync_in_background(watchdog)

    def __del__(self):
        if not hasattr(self, "_features"):
            # not initialized yet
            return
        self.set_feature("watch_update", FeatureConfig(enabled=False, args={}))

    def close(self):
        if not hasattr(self, "_features"):
            # not initialized yet
            return
        self.set_feature("watch_update", FeatureConfig(enabled=False, args={}))


class RunchAsyncCustomConfigReader[C: RunchModel, *Ts]:
    _config_name: str
    _config_encoding: str

    _config_schema: Type[C]
    _config: Runch[C] | None
    _config_updated_at: datetime | None

    _config_loader: Callable[[str, *Ts], Coroutine[Any, Any, C]]
    _config_loader_extra_args: tuple[*Ts]

    _features: MutableMapping[AsyncFeatureKey, FeatureConfig]
    _auto_update_started: bool

    if TYPE_CHECKING:
        # this attribute is defined on the Generic class, simply adding it here to avoid pyright error
        __orig_class__: Type[RunchConfigReader[C]]

    def __init__(
        self,
        config_name: str,
        config_encoding: str = "utf-8",
        *,
        config_loader: Callable[[str, *Ts], Coroutine[Any, Any, C]],
        config_loader_extra_args: tuple[*Ts] = (),
        features: Mapping[AsyncFeatureKey, FeatureConfig] | None = None,
    ):
        orig_class: type[Any] = get_orig_class(self)
        if len(get_args(orig_class)) < 2:
            # RunchAsyncCustomConfigReader[C] -> RunchAsyncCustomConfigReader[C, ()]
            orig_class = RunchAsyncCustomConfigReader[get_args(orig_class)[0], ()]

        self._config = None
        self._config_updated_at = None
        self._config_name = config_name
        self._config_encoding = config_encoding
        self._config_schema = get_generic_arg_kv_map(orig_class)[C]
        self._config_loader = config_loader
        self._config_loader_extra_args = config_loader_extra_args
        self._features = {}
        self._auto_update_started = False

        if features is not None:
            for feature, value in features.items():
                self.set_feature(feature, value)

        # make sure _features is always fully initialized
        for feature, value in _CUSTOM_A_CONFIG_READER_DEFAULT_FEATURES.items():
            if feature not in self._features:
                self.set_feature(feature, value)

    async def read(self) -> Runch[C]:
        if self._config is not None:
            return self._config

        type_ = self._config_schema

        config = await self._config_loader(
            self._config_name, *self._config_loader_extra_args
        )

        if (
            self._config_updated_at is None
            and self._features["watch_update"]["enabled"]
            and not self._auto_update_started
        ):
            self._start_auto_update()

        self._config = Runch[type_](config)
        self._config_updated_at = datetime.now()

        return self._config

    async def update(
        self,
        *,
        overwrite_uninitialized: bool = False,
        on_error: Literal["ignore", "raise"] = "ignore",
    ):
        """This function will try to update the config base on the latest config file.

        Args:
            overwrite_uninitialized (bool, optional): _If set to False, then update() will become a noop before read() is ever called. This can have side effects when used together with `read_lazy()`._
            Defaults to `False`.

            on_error ("raise" | "ignore", optional): _Specifies the behavior when errors occurred during an attempt to update this config._
            Defaults to `"ignore"`, which means ignore the error and keep the old config.
        """

        if self._config is None:
            if self._config_updated_at is not None:
                raise RuntimeError(
                    "update: _config is None but _config_updated_at is not None. This is a bug."
                )
            if overwrite_uninitialized:
                # force update even if the config is not initialized due to lazy load
                try:
                    await self.read()
                except Exception:
                    if on_error == "raise":
                        raise

            return

        type_ = self._config_schema

        try:
            new_config = Runch[type_](
                await self._config_loader(
                    self._config_name, *self._config_loader_extra_args
                )
            )
        except Exception:
            if on_error == "raise":
                raise
            else:
                return

        self._config.update(new_config)
        self._config_updated_at = datetime.now()

    def set_feature(
        self, feature: AsyncFeatureKey, feature_config: FeatureConfig
    ) -> Self:
        """Set a feature's configuration. See **Features** section for available features.

        Args:
            feature (AsyncFeatureKey): The feature's key
            value (FeatureConfig): The feature's configuration.
              - enabled: Whether the feature is enabled
              - args: The arguments for the feature

        ## Features
        - feature="watch_update": Automatically update the config every `n` seconds.
            - args: `{"update_interval": n, "on_update_error": "ignore" | "raise"}`, where `update_interval` is always required.
        """
        self._features[feature] = feature_config

        if (
            feature == "watch_update"
            and feature_config["enabled"]
            and self._config_updated_at is not None
        ):
            self._start_auto_update()

        return self

    def enable_feature(self, feature: AsyncFeatureKey, args: dict[str, Any]) -> Self:
        return self.set_feature(feature, FeatureConfig(enabled=True, args=args))

    def disable_feature(self, feature: AsyncFeatureKey) -> Self:
        return self.set_feature(feature, FeatureConfig(enabled=False, args={}))

    def _start_auto_update(self):
        if not self._features["watch_update"]["enabled"]:
            return

        if "update_interval" not in self._features["watch_update"]["args"]:
            raise ValueError("watch_update feature requires `update_interval` in args")

        if "on_update_error" in self._features["watch_update"]["args"]:
            on_update_error = self._features["watch_update"]["args"]["on_update_error"]
            if on_update_error not in ["ignore", "raise"]:
                raise ValueError('on_update_error must be either "ignore" or "raise"')
        else:
            # will fallback to default later
            on_update_error = None

        if self._auto_update_started:
            return

        self._auto_update_started = True

        # use weakref to avoid circular reference to `self`
        self_ref = weakref.ref(self)

        # watchdog will stop automatically if the feature is disabled / the reader is deleted
        async def a_watchdog():
            # NOTE: under very rare circumstances, the auto update watchdog thread may not be able to be recreated.
            # That is, if the user disables this feature, watchdog will quit in the next iteration. However, if the
            # user enables the feature again between the `break` of `while True` and `self._auto_updating = False`,
            # the new watchdog thread will not be created due to the `if self._auto_updating` check.
            # This is a very very rare edge case and is not quite fixable without adding a lock and some complexity.
            while True:
                self_ = self_ref()
                if self_ is None:
                    # reader is deleted
                    break

                await asyncio.sleep(
                    self_._features["watch_update"]["args"]["update_interval"]
                )

                if not self_._features["watch_update"]["enabled"]:
                    # auto update is turned off
                    break

                if on_update_error is not None:
                    await self_.update(
                        overwrite_uninitialized=False, on_error=on_update_error
                    )
                else:
                    await self_.update(overwrite_uninitialized=False)

                del self_

            self_ = self_ref()
            if self_ is not None:
                self_._auto_update_started = False

        def watchdog():
            asyncio.run(a_watchdog())

        _run_sync_in_background(watchdog)

    def __del__(self):
        if not hasattr(self, "_features"):
            # not initialized yet
            return
        self.set_feature("watch_update", FeatureConfig(enabled=False, args={}))

    def close(self):
        if not hasattr(self, "_features"):
            # not initialized yet
            return
        self.set_feature("watch_update", FeatureConfig(enabled=False, args={}))


def require_lazy_runch_configs(*runches: Runch[Any]):
    """
    This function is used to force evaluation of runch configs, useful if you are using read_lazy().
    """
    for runch in runches:
        # access any attribute (e.g. __module__) to force evaluation.
        runch.config.__module__



import sys

from runch.script.schema_generator import (
    generate_model,
    __doc__ as schema_generator_doc,
)
from runch._reader import _SupportedFileType

from typing import cast

if __name__ == "__main__":
    # TODO: move to argparse

    if len(sys.argv) < 2:
        print(
            schema_generator_doc,
            file=sys.stderr,
        )
        sys.exit(1)

    config_path = sys.argv[1]
    config_name = None
    config_ext: _SupportedFileType = "yaml"

    if len(sys.argv) == 3:
        config_name = sys.argv[2]
    elif len(sys.argv) == 4:
        config_name = sys.argv[2]
        config_ext = cast(_SupportedFileType, sys.argv[3])
    elif len(sys.argv) > 4:
        print(
            __doc__,
            file=sys.stderr,
        )
        sys.exit(1)

    model = generate_model(config_path, config_ext, config_name=config_name)
    print(model)



# pyright: basic
import os
import re

from datamodel_code_generator.format import CustomCodeFormatter
from typing import Any

CONFIG_BODY_IDENT = "# config_body_start"
CAPITALIZE_WORDS = {
    "mp3",
    "pdf",
    "json",
    "lb",
    "url",
    "uri",
    "id",
    "ip",
    "db",
    "obs",
    "sms",
    "cn",
    "en",
    "zh",
    "sso",
    "html",
    "ivs",
}


def camel_to_splitted(name: str) -> list[str]:
    return (
        "".join([f"_{char.lower()}" if char.isupper() else char for char in name])
        .lstrip("_")
        .split("_")
    )


def snake_to_camel(name: str) -> str:
    splitted = [word for word in name.split("_")]
    for i, word in enumerate(splitted):
        if len(word) <= 2 or word.lower() in CAPITALIZE_WORDS:
            splitted[i] = word.upper()
        else:
            splitted[i] = word.capitalize()

    return "".join(splitted)


pydantic_import_re = re.compile(
    r"from pydantic import ([\w, ]+?)?(?:, )?BaseModel(?:, )?([\w, ]+)?"
)


class CodeFormatter(CustomCodeFormatter):
    config_file_ext: str
    config_name: str
    config_path: str
    config_type: str

    def __init__(self, formatter_kwargs: dict[str, Any]) -> None:
        super().__init__(formatter_kwargs)
        self.config_file_ext = formatter_kwargs["config_file_ext"]
        self.config_name = formatter_kwargs["config_name"]
        self.config_path = formatter_kwargs["config_path"]
        self.config_type = formatter_kwargs["config_type"]

    def apply(self, code: str) -> str:
        config_class_name = snake_to_camel(self.config_name)

        pydantic_import_match = pydantic_import_re.search(code)
        assert (
            pydantic_import_match is not None
        ), "datamodel-code-generator created unexpected result"

        pydantic_import_left, pydantic_import_right = pydantic_import_match.groups()

        code = pydantic_import_re.sub("from pydantic import BaseModel", code)

        if pydantic_import_left and pydantic_import_right:
            pydantic_import = (
                f"from pydantic import {pydantic_import_left}, {pydantic_import_right}"
            )
        elif pydantic_import_left:
            pydantic_import = f"from pydantic import {pydantic_import_left}"
        elif pydantic_import_right:
            pydantic_import = f"from pydantic import {pydantic_import_right}"
        else:
            pydantic_import = None

        new_imports = [
            f"from runch import RunchModel, RunchConfigReader",
            CONFIG_BODY_IDENT,
        ]

        if pydantic_import is not None:
            new_imports.insert(0, pydantic_import)

        code = code.replace(
            "from pydantic import BaseModel",
            "\n".join(new_imports),
        )
        code = code.replace("class Model(", f"class {config_class_name}ConfigModel(")
        code = code.replace("(BaseModel):", "(RunchModel):")

        config_head_start = code.find(
            f"class {config_class_name}ConfigModel(RunchModel):"
        )
        config_head = code[config_head_start:].strip("\n")

        config_body_start = code.find(CONFIG_BODY_IDENT)
        config_body = code[
            config_body_start + len(CONFIG_BODY_IDENT) : config_head_start
        ].strip("\n")

        config_imports = code[:config_body_start].strip("\n")

        config_renames: dict[str, str] = {}

        def transform_body_line(line: str) -> str:
            if line.startswith("class "):
                class_name = line.split("(")[0].split("class ")[1]
                class_name_splitted = camel_to_splitted(class_name)
                for i, word in enumerate(class_name_splitted):
                    if len(word) <= 2 or word in CAPITALIZE_WORDS:
                        class_name_splitted[i] = word.upper()
                    else:
                        class_name_splitted[i] = word.capitalize()

                # handle schema deduplication suffix generated by datamodel-codegen
                schema_dedup_id: str | None = None
                if re.match(r"^[A-Za-z]+\d+$", class_name_splitted[-1]):
                    _schema_dedup_id: str = re.findall(
                        r"\d+$", class_name_splitted[-1]
                    )[0]
                    schema_dedup_id = _schema_dedup_id
                    class_name_splitted[-1] = class_name_splitted[-1].replace(
                        _schema_dedup_id, f""
                    )

                new_class_name = "".join(class_name_splitted)
                config_renames[class_name] = new_class_name + "Config"
                if schema_dedup_id:
                    config_renames[class_name] += schema_dedup_id

            if line.strip() == "":
                return ""
            else:
                return line

        # add indent
        config_body = "\n".join(map(transform_body_line, config_body.split("\n")))
        config_classes = f"{config_head}\n\n{config_body}"

        for old_name, new_name in config_renames.items():
            old_name_pattern_1 = re.compile(rf" {old_name}\b")
            old_name_pattern_2 = re.compile(rf"\[{old_name}\b")
            config_classes = old_name_pattern_1.sub(f" {new_name}", config_classes)
            config_classes = old_name_pattern_2.sub(f"[{new_name}", config_classes)

        config_file_fullname = os.path.basename(self.config_path)
        config_file_path = os.path.dirname(self.config_path)

        config_read_directive = f'_{self.config_name}_reader = RunchConfigReader[{config_class_name}ConfigModel]("{config_file_fullname}", config_dir="{config_file_path}", config_type="{self.config_type}")'
        config_read_directive += (
            f"\n{self.config_name} = _{self.config_name}_reader.read_lazy()"
        )
        config_read_directive += (
            "\n\n# uncomment the following line to enable the watch_update feature"
        )
        config_read_directive += (
            f"\n# _{self.config_name}_reader"
            + '.set_feature("watch_update", {"enabled": True, "args": {"update_interval": 10}})'
        )

        return f"{config_imports}\n\n{config_classes}\n\n{config_read_directive}"



from time import sleep
from runch import (
    RunchModel,
    RunchConfigReader,
)


class TestConfig(RunchModel):
    x: int


test_reader = RunchConfigReader[TestConfig](
    config_name="test.yaml", config_dir="runch/test"
)

test_reader.enable_feature("watch_update", {"update_interval": 1})

test_config = test_reader.read_lazy()

while True:
    print(test_config.config)
    sleep(1)



import asyncio
import random

from httpx import AsyncClient

from runch import (
    RunchModel,
    RunchAsyncCustomConfigReader,
)


class TestConfig(RunchModel):
    status: str
    method: str


async def config_loader(config_name: str, auth: str | None = None) -> TestConfig:
    """Load config from a remote source."""

    print(f"Loading config from remote source for {config_name=}...")

    headers = {"Authorization": f"Bearer {auth}"}

    # Simulate a network request to fetch the config.

    async with AsyncClient() as client:
        if random.random() < 0.5:
            response = await client.get(
                "https://dummyjson.com/test",
                headers=headers,
            )
        else:
            response = await client.post(
                "https://dummyjson.com/test",
                headers=headers,
            )

        response.raise_for_status()

    return TestConfig(**response.json())


test_reader_1 = RunchAsyncCustomConfigReader[TestConfig](
    config_name="example1",
    config_loader=config_loader,
)

test_reader_2 = RunchAsyncCustomConfigReader[TestConfig, *tuple[str]](
    config_name="example2",
    config_loader=config_loader,
    config_loader_extra_args=("sk-example_secret_auth_key",),
)

test_reader_1.enable_feature("watch_update", {"update_interval": 2})
test_reader_2.enable_feature("watch_update", {"update_interval": 2})


async def main():
    test_config_1 = await test_reader_1.read()
    test_config_2 = await test_reader_2.read()

    while True:
        print("test_config_1.config", test_config_1.config)
        print("test_config_2.config", test_config_2.config)
        await asyncio.sleep(2)


asyncio.run(main())



from pathlib import Path
from tempfile import TemporaryDirectory
from datamodel_code_generator import DataModelType, InputFileType, generate

with TemporaryDirectory() as temporary_directory_name:
    temporary_directory = Path(temporary_directory_name)
    output = Path(temporary_directory / "model.py")
    generate(
        {"foo": 1, "bar": {"baz": 2}},
        input_file_type=InputFileType.Dict,
        input_filename="placeholder",
        output=output,
        output_model_type=DataModelType.PydanticV2BaseModel,
        snake_case_field=True,
    )
    model: str = output.read_text()

print(model)
